unit NkVImage;

interface

uses Windows, SUtil, SimpleList, SPIUtil, DIBLib;

type
  PMaxBitmapInfo = ^TMaxBitmapInfo;
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    case Boolean of
      //True: (bmiColors: array[0..0] of TRGBQuad);
      True: (bmiColors: array[Byte] of TRGBQuad);
      False: (bmiColorsBF: array[0..2] of DWORD);
  end;

  TSPIError = record
    Code: Integer;
    Handle: THandle;
  end;

  TNkVImage = class(TObject)
  private
    SPIHBm: HLOCAL; // SPI pnh
    function FindSPI(const FileName: string; Buf: PChar; var PluginHandle: THandle): Integer;
    function GetImage(PluginHandle: THandle; pSrc: PChar; SrcSize: Integer): Integer;
  public
    PluginList: TSimpleList;
    Loaded: Boolean;
    Width: Integer;
    Height: Integer;
    ColorDepth: Integer;
    FileSize: Integer;
    FileName: string;
    FileCTime: TFileTime;       // t@C쐬
    FileATime: TFileTime;       // t@CŏIANZX
    FileMTime: TFileTime;       // t@CŏIݓ
    MemSize: Integer;
    Format: string;             // tH[}bg
    SPIName: string;            // Susie plug-in 
    SPIFileName: string;        // Susie plug-in t@C
    Memo: string;               // eLXg
    BitmapInfo: TMaxBitmapInfo;
    pBits: Pointer;             // rbgf[^
    SPIError: TSPIError;        // Susie plug-inG[R[h
    hAM: THandle;               // A[JCut@ĈƂ͔0
    ArcFileName: string;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function LoadFromFile(const FileName: string): Integer;
    function LoadFromArchive(ArcFileName: string; const FileItem: PFileItem): Integer;
    procedure FreeAM;
  end;

const
  nilfUnknown = -1;
  nilfArchive = -2;

  HeadBufSize = 2048; // tH[}bĝ߂ɓǂݍbyte

implementation

function CountBit(Value: Integer): Integer;
begin
  Result := 0;
  repeat
    Inc(Result, Value and 1);
    Value := Value shr 1
  until Value = 0;
end;

function ToCRLF(const S: string): string;
var
  CRflg: Boolean;
  i, len: Integer;
begin
  CRflg := False;
  Result := '';
  len := Length(S);
  for i := 1 to len do
  begin
    if CRflg and (S[i] <> #10) then
      Result := Result + #10
    else if (not CRflg) and (S[i] = #10) then
      Result := Result + #13;
    Result := Result + S[i];
    CRflg := (S[i] = #13);
  end;
end;

function GetBmpFormatStr(pBmi: PBitmapInfo): string;
var
  s: string;
begin
  if pBmi.bmiHeader.biSize = SizeOf(TBitmapInfoHeader) then
  begin
    case pBmi.bmiHeader.biCompression of
      BI_RGB:  s := 'RGB';
      BI_RLE8: s := 'RLE8';
      BI_RLE4: s := 'RLE4';
      BI_BITFIELDS:
        begin
          s := 'BITFIELDS';
          with PMaxBitmapInfo(pBmi)^ do
          begin
            s := s + ' ('+IntToStr(CountBit(bmiColorsBF[0]))+
              IntToStr(CountBit(bmiColorsBF[1]))+
              IntToStr(CountBit(bmiColorsBF[2]))+')';
          end;
        end;
    else
      s := 'unknown';
    end;
    if pBmi.bmiHeader.biHeight < 0 then
      s := s + ' (topdown)';
    Result := 'Windows Bitmap - ' + s;
  end
  else if pBmi.bmiHeader.biSize = SizeOf(TBitmapCoreHeader) then
    Result := 'OS/2 Bitmap'
  else
    Result := '';
end;

// pBuffer: 2KBytes炢͂ق
function GetFormatStr(pBuffer: PChar): string;
begin
  Result := '';
  // BMP
  if Copy(pBuffer, 1, 2) = 'BM' then
    Result := GetBmpFormatStr(@pBuffer[SizeOf(TBitmapFileHeader)])
  // GIF
  else if Copy(pBuffer, 1, 3) = 'GIF' then
    Result := Copy(pBuffer, 1, 6)
  // JPEG(JFIF)
  else if (Copy(pBuffer, 1, 2) = #$FF#$D8) and (Copy(PChar(@pBuffer[6]), 1, 4) = 'JFIF') then
    Result := 'JPEG - JFIF'
  // Exif?
  else if (Copy(pBuffer, 1, 2) = #$FF#$D8) and (Copy(PChar(@pBuffer[6]), 1, 4) = 'Exif') then
    Result := 'Exif'
  // PNG
  else if Copy(pBuffer, 1, 8) = #$89'PNG'#$0D#$0A#$1A#$0A then
  begin
    Result := 'PNG';
    if pBuffer[28] = #0 then
      Result := Result + ' - non-interlaced'
    else if pBuffer[28] = #1 then
      Result := Result + ' - interlaced';
    if Ord(pBuffer[25]) and 4 <> 0 then
      Result := Result + ' (with alpha)';
  end;
end;

{ TNkVImage }

constructor TNkVImage.Create;
begin
  PluginList := TSimpleList.Create;
end;

destructor TNkVImage.Destroy;
var
  i: Integer;
begin
  Clear;
  for i := 0 to PluginList.Count - 1 do
    FreeMem(PluginList[i]);
  PluginList.Free;
  if hAM <> 0 then
    FreeLibrary(hAM);
end;

procedure TNkVImage.Clear;
begin
  if SPIHBm = 0 then
  begin
    if pBits <> nil then
      GlobalFree(HGLOBAL(pBits))
  end
  else begin
    LocalUnLock(SPIHBm);
    LocalFree(SPIHBm);
    SPIHBm := 0;
  end;
  pBits := nil;
  Loaded := False;
  Memo := '';
  SPIName := '';
  SPIFileName := '';
end;

function TNkVImage.FindSPI(const FileName: string; Buf: PChar; var PluginHandle: THandle): Integer;
var
  i: Integer;
begin
  PluginHandle := 0;
  Result := GF_UNKNOWN;

  if (Buf[0] = 'B') and (Buf[1] = 'M') then
  begin
    Result := GF_IMAGE;
    Exit;
  end;

  // SPI
  for i := 0 to PluginList.Count - 1 do
  begin
    Result := GetFileFormat(PluginList[i], FileName, Buf, PluginHandle);
    if Result <> GF_UNKNOWN then
      Exit;
  end;
end;

function TNkVImage.GetImage(PluginHandle: THandle; pSrc: PChar; SrcSize: Integer): Integer;
var
  BitmapInfo: TMaxBitmapInfo;
  pDst: Pointer;
  DstSize: Integer;
  ColorDepth: Integer;
  TextInfo: string;
  HBm: HLOCAL;
  buf: array[0..MAX_PATH] of Char;
begin
  Result := 0;
  if PluginHandle = 0 then
  begin
    // DIB
    HBm := 0;
    ExtractDIB(pSrc, SrcSize, PBitmapInfo(@BitmapInfo), pDst, DstSize, ColorDepth, TextInfo);
  end else
    // SPI Plug-in
    Result := ExtractImage(PluginHandle, pSrc, SrcSize, PBitmapInfo(@BitmapInfo), pDst, DstSize,
      ColorDepth, TextInfo, HBm);
  // test
  //Result := 8;
  if Result <> 0 then
    Exit;

  Clear;
  Loaded := True;
  SPIHBm := HBm;
  MemSize := DstSize;
  Self.BitmapInfo := BitmapInfo;
  Width := BitmapInfo.bmiHeader.biWidth;
  Height := BitmapInfo.bmiHeader.biHeight;
  pBits := pDst;
  Self.ColorDepth := ColorDepth;
  Memo := ToCRLF(TextInfo);
  Format := GetFormatStr(pSrc);
  FileSize := SrcSize;
  if PluginHandle <> 0 then
  begin
    GetModuleFileName(PluginHandle, buf, SizeOf(buf));
    SPIFileName := buf;
    SPIName := GetPluginName(PluginHandle);
  end else
    SPIName := '(Internal DIB to DIB filter)';
end;

function TNkVImage.LoadFromFile(const FileName: string): Integer;
var
  pSrc: Pointer;
  h: THandle;
  F: File;
  size: Integer;
  PluginHandle: THandle;
  fd: TWin32FindData;
  r: Integer;
  HeadBuf: array[0..HeadBufSize-1] of Char;
  fmt: Integer;
begin
  pSrc := nil;

  // open O(ANZX^CXVO) Find Data 擾
  h := FindFirstFile(PChar(FileName), fd);
  if h <> INVALID_HANDLE_VALUE then
    Windows.FindClose(h);
  AssignFile(F, FileName);
  {$I-}
  Reset(F, 1);
  {$I+}
  Result := IOResult;
  if Result <> 0 then
    Exit;
  size := System.FileSize(F);
  FillChar(HeadBuf, SizeOf(HeadBuf), 0);
  BlockRead(F, HeadBuf, SizeOf(HeadBuf), r);
  // fmt = GF_IMAGEPluginHandle=0Ȃbmp
  fmt := FindSPI(FileName, HeadBuf, PluginHandle);
  if fmt = GF_IMAGE then
  begin
    GetMem(pSrc, size);
    Seek(F, 0);
    BlockRead(F, pSrc^, size);
  end;
  CloseFile(F);

  case fmt of
    GF_UNKNOWN:
      Result := nilfUnknown;
    GF_ARCHIVE:
      begin
        if hAM <> 0 then
          FreeLibrary(hAM);
        hAM := PluginHandle;
        Result := nilfArchive;
      end;
    GF_IMAGE:
      begin
        SPIError.Code := GetImage(PluginHandle, pSrc, size);
        //SPIError.Code := 3; // error test
        SPIError.Handle := PluginHandle;
        FreeMem(pSrc);
        if SPIError.Code <> 0 then
        begin
          Result := nilfUnknown;
          Exit;
        end;
        Self.FileName := FileName;
        FileCTime := fd.ftCreationTime;
        FileMTime := fd.ftLastWriteTime;
        FileATime := fd.ftLastAccessTime;
        if PluginHandle <> 0 then
          FreeLibrary(PluginHandle);
        if hAM <> 0 then
        begin
          FreeLibrary(hAM);
          hAM := 0;
        end;
      end;
  end;
end;

function TNkVImage.LoadFromArchive(ArcFileName: string; const FileItem: PFileItem): Integer;
var
  pTmp: PChar;
  size: Integer;
  PluginHandle: THandle;
  HeadBuf: array[0..HeadBufSize-1] of Char;
  fmt: Integer;
begin
  Result := 0;
  if hAM = 0 then
  begin
    Result := -1;
    Exit;
  end;
  if ArcFileName = '' then
    ArcFileName := Self.ArcFileName;
  SPIError.Code := GetFileFromArchive(hAM, ArcFileName, FileItem.Position, pTmp, size);
  SPIError.Handle := hAM;
  if SPIError.Code <> 0 then
  begin
    Result := nilfUnknown;
    Exit;
  end;
  if size < SizeOf(HeadBuf) then
    Move(pTmp^, HeadBuf, size)
  else
    Move(pTmp^, HeadBuf, SizeOf(HeadBuf));
  fmt := FindSPI(FileItem.FileName, HeadBuf, PluginHandle);
  case fmt of
    GF_UNKNOWN:
      Result := nilfUnknown;
    GF_ARCHIVE:
      begin
        if hAM <> 0 then  // A[JCuA[JCu͖Ή
          FreeLibrary(PluginHandle);
        Result := nilfUnknown;
      end;
    GF_IMAGE:
      begin
        SPIError.Code := GetImage(PluginHandle, pTmp, size);
        SPIError.Handle := PluginHandle;
        if SPIError.Code <> 0 then
        begin
          Result := nilfUnknown;
          Exit;
        end;
        Self.ArcFileName := ArcFileName;
        FileName := FileItem.FileName;
        FillChar(FileCTime, SizeOf(FileCTime) ,0);
        FileMTime := FileItem.FileMTime;
        FillChar(FileATime, SizeOf(FileATime) ,0);
        if PluginHandle <> 0 then
          FreeLibrary(PluginHandle);
      end;
  end;
  FreeMem(pTmp);
end;

procedure TNkVImage.FreeAM;
begin
  if hAM <> 0 then
  begin
    FreeLibrary(hAM);
    hAM := 0;
  end;
end;

end.
