unit SMenu;

interface

uses Windows;

type
  TSMenu = class(TObject)
  private
    FHandle: HMENU;
  public
    property Handle: HMENU read FHandle;
    procedure AppendItem(Id: Integer; Caption: PChar);
    procedure InsertItem(Id: Integer; Index: Integer; Caption: PChar);
    procedure GrayedItem(Id: Integer);
    procedure GrayedItemByPos(Index: Integer);
    procedure CheckItem(Id: Integer; Checked: Boolean);
    function IndexOfItem(Id: Integer): Integer;
    constructor Create; virtual;
    destructor Destroy; override;
  end;

  TSPopupMenu = class(TSMenu)
    constructor Create; override;
    procedure Popup(OwnerHandle: HWND);
  end;

implementation

{ TSMenu }

constructor TSMenu.Create;
begin
  FHandle := CreateMenu;
end;

destructor TSMenu.Destroy;
begin
  DestroyMenu(FHandle);
  inherited Destroy;
end;

procedure TSMenu.AppendItem(Id: Integer; Caption: PChar);
var
  flag: Integer;
begin
  if IsMenu(Id) then
    flag := MF_POPUP
  else
    flag := MF_STRING;
  if Caption = '-' then
    flag := flag or MF_SEPARATOR;
  AppendMenu(FHandle, flag, Id, Caption);
end;

procedure TSMenu.InsertItem(Id: Integer; Index: Integer; Caption: PChar);
var
  flag: Integer;
begin
  if IsMenu(Id) then
    flag := MF_POPUP
  else
    flag := MF_STRING;
  if Caption = '-' then
    flag := flag or MF_SEPARATOR;
  InsertMenu(FHandle, Index, flag or MF_BYPOSITION, Id, Caption);
end;

procedure TSMenu.GrayedItem(Id: Integer);
begin
  EnableMenuItem(FHandle, Id, MF_GRAYED);
end;

procedure TSMenu.GrayedItemByPos(Index: Integer);
begin
  EnableMenuItem(FHandle, Index, MF_GRAYED or MF_BYPOSITION);
end;

procedure TSMenu.CheckItem(Id: Integer; Checked: Boolean);
begin
  if Checked then
    CheckMenuItem(FHandle, Id, MF_CHECKED)
  else
    CheckMenuItem(FHandle, Id, MF_UNCHECKED);
end;

function TSMenu.IndexOfItem(Id: Integer): Integer;
var
  i, cnt: Integer;
begin
  cnt := GetMenuItemCount(FHandle);
  for i := 0 to cnt - 1 do
  begin
    if GetMenuItemID(FHandle, i) = Id then
    begin
      Result := i;
      Exit;
    end;
  end;
  Result := -1;
end;

{ TSPopupMenu }

constructor TSPopupMenu.Create;
begin
  FHandle := CreatePopupMenu;
end;

procedure TSPopupMenu.Popup(OwnerHandle: HWND);
var
  pt: TPoint;
begin
  GetCursorPos(pt);
  TrackPopupMenu(FHandle, TPM_LEFTALIGN or TPM_LEFTBUTTON or TPM_RIGHTBUTTON,
    pt.x, pt.y, 0, OwnerHandle, nil);
end;

end.
