{
  Susie Plugin Helper Unit
  2002/11/10-
}

unit SPIUtil;

interface

uses Windows, spi_api;

const
  GF_UNKNOWN = 0;
  GF_IMAGE = 1;
  GF_ARCHIVE = 2;

  SpiErrMsg: array[0..10] of PChar = (
    '',
    'user cancel',
    'unknown format',
    'broken data',
    'allocate memory error',
    'memory error',
    'file read error',
    'cannot create window',
    'internal error',
    'file write error',
    'end of file error'
  );

type
  PFileItem = ^TFileItem;
  TFileItem = record
    FileName: array[0..MAX_PATH] of Char;
    FileMTime: TFileTime;
    IsImage: Boolean;
    Position: Integer; // for archive
  end;

function GetPluginPath: string;
function ExtractImage(PluginHandle: THandle; pSrc: PChar; SrcSize: Integer;
  BitmapInfo: PBitmapInfo; var pDst: Pointer; var DstSize: Integer;
  var ColorDepth: Integer; var TextInfo: string; var HBm: HLOCAL): Integer;
function GetPluginName(PluginHandle: THandle): string;
function GetFileListFromArchive(PluginHandle: THandle; const FileName: string;
  var pBuf: PFileItem): Integer;
function GetFileFromArchive(PluginHandle: THandle; const FileName: string;
  Position: Integer; var pBuf: PChar; var size: Integer): Integer;
function GetFileFormat(PluginFileName: PChar; const FileName: string; pBuf: PChar; var PluginHandle: THandle): Integer;
procedure ConfigDlg(hWnd: HWND; FileName: PChar);

implementation

// Susie plug-in ̏ꏊ擾
function GetPluginPath: string;
var
  data: array[0..MAX_PATH] of Char;
  size: DWORD;
  hkResult: HKEY;
begin
  FillChar(data, SizeOf(data), 0);
  if RegOpenKeyEx(HKEY_CURRENT_USER, 'Software\Takechin\Susie\Plug-in', 0,
    KEY_QUERY_VALUE, hkResult) = ERROR_SUCCESS then
  begin
    size := SizeOf(data);
    RegQueryValueEx(hkResult, 'Path', nil, nil, @data, @size);
    RegCloseKey(hkResult);
  end;
  Result := data;
end;

// time_tFileTimeɕϊ
function time_tToFileTime(t: Longint): TFileTime;
var
  st: TSystemTime;
  ft: TFileTime;
begin
  st.wYear := 1970;
  st.wMonth := 1;
  st.wDay := 1;
  st.wHour := 0;
  st.wMinute := 0;
  st.wSecond := 0;
  st.wMilliseconds := 0;
  SystemTimeToFileTime(st, ft);
  Comp(Result) := Comp(ft) + t * 10000000.0;
end;

function GetFileListFromArchive(PluginHandle: THandle; const FileName: string;
  var pBuf: PFileItem): Integer;
var
  GetArchiveInfo: TGetArchiveInfo;
  hinf: HLOCAL;
  p: PFileInfo;
  pi: PFileItem;
  cnt: Integer;
  s: string;
begin
  GetArchiveInfo := GetProcAddress(PluginHandle, 'GetArchiveInfo');
  Result := GetArchiveInfo(PChar(FileName), 0, 0, hinf);
  // lhasad.spi 0.7͐ɂ炸unknown format(=2)ԂĂ
  // IsSupportedɐĂȏAunknown format̉\͒ႢƂƂ
  // I0ɏ
  if Result = 2 then
    Result := 0;
  if Result <> 0 then
    Exit;
  p := LocalLock(hinf);
  cnt := LocalSize(hinf) div SizeOf(TFileInfo) + 1;
  if p = nil then
  begin
    pBuf := nil;
    Exit;
  end;
  GetMem(pBuf, SizeOf(TFileItem) * cnt);
  FillChar(pBuf^, SizeOf(TFileItem) * cnt, 0);
  pi := pBuf;
  while p.Method[0] <> #0 do
  begin
    if p.FileSize > 0 then // FileSize = 0 ̓fBNg
    begin
      s := p.Path + p.FileName;
      lstrcpyn(pi.FileName, PChar(s), SizeOf(pi.FileName));
      pi.FileMTime := time_tToFileTime(p.TimeStamp);
      pi.IsImage := True;
      pi.Position := p.Position;
      Inc(pi);
    end;
    Inc(p);
  end;
  LocalUnlock(hinf);
  LocalFree(hinf);
end;

function GetFileFromArchive(PluginHandle: THandle; const FileName: string;
  Position: Integer; var pBuf: PChar; var size: Integer): Integer;
var
  GetFile: TGetFile;
  HDst: HLOCAL;
  p: PChar;
begin
  GetFile := GetProcAddress(PluginHandle, 'GetFile');
  Result := GetFile(PChar(FileName), Position, @HDst, $100, nil, 0);
  // test
  //Result := 10;
  if Result <> 0 then
    Exit;
  p := LocalLock(HDst);
  size := LocalSize(HDst);
  GetMem(pBuf, size);
  Move(p^, pBuf^, size);
  LocalUnLock(HDst);
  LocalFree(HDst);
end;

procedure ConfigDlg(hWnd: HWND; FileName: PChar);
var
  h: THandle;
  ConfDlg: TConfigurationDlg;
begin
  h := LoadLibrary(FileName);
  if h = 0 then
    Exit;
  ConfDlg := GetProcAddress(h, 'ConfigurationDlg');
  if Assigned(ConfDlg) then
  begin
    ConfDlg(hWnd, 1);
  end else
    MessageBox(hWnd, 'Configuration Dialog is not implemented.', 'Plug-in Setting',
      MB_OK or MB_ICONINFORMATION);
end;

function ProgressCallback(nNum, nDemon: Integer; lData : Longint): Integer; stdcall;
begin
  Result := 0;
end;

function GetPluginName(PluginHandle: THandle): string;
var
  GetPluginInfo: TGetPluginInfo;
  buf: array[0..1024] of Char;
begin
  GetPluginInfo := GetProcAddress(PluginHandle, 'GetPluginInfo');
  GetPluginInfo(1, @buf, SizeOf(buf));
  Result := buf;
end;

function ExtractImage(PluginHandle: THandle; pSrc: PChar; SrcSize: Integer;
  BitmapInfo: PBitmapInfo; var pDst: Pointer; var DstSize: Integer;
  var ColorDepth: Integer; var TextInfo: string; var HBm: HLOCAL): Integer;
var
  GetPicture: TGetPicture;
  GetPictureInfo: TGetPictureInfo;
  PicInfo: TPictureInfo;
  HBInfo: HLOCAL;
  p: PChar;
begin
  TextInfo := '';
  if PluginHandle = 0 then
  begin
    Result := -1;
    Exit;
  end;

  GetPictureInfo := GetProcAddress(PluginHandle, 'GetPictureInfo');

  // PicInfoSZbgȂvOC΍
  FillChar(PicInfo, SizeOf(PicInfo), 0);
  Result := GetPictureInfo(pSrc, SrcSize, 1, PicInfo);
  if Result <> 0 then
    Exit;
  ColorDepth := PicInfo.colorDepth;
  if PicInfo.hInfo <> 0 then
  begin
    p := LocalLock(PicInfo.hInfo);
    SetLength(TextInfo, Length(p));
    Move(p^, PChar(TextInfo)^, Length(p));
    LocalUnLock(PicInfo.hInfo);
    LocalFree(PicInfo.hInfo);
  end;

  GetPicture := GetProcAddress(PluginHandle, 'GetPicture');
  // ProgressCallbacknilƗvOĈŃ_~[֐n
  Result := GetPicture(pSrc, SrcSize, 1, HBInfo, HBm, ProgressCallback, 0);
  if Result = 0 then
  begin
    if HBInfo <> 0 then
      Move((LocalLock(HBInfo))^, BitmapInfo^, LocalSize(HBInfo));
    LocalUnLock(HBInfo);
    LocalFree(HBInfo);
    pDst := LocalLock(HBm);
    DstSize := LocalSize(HBm);
  end;
end;

// GF_UNKNOWN̂Ƃ̂FreeLibaray
function GetFileFormat(PluginFileName: PChar; const FileName: string; pBuf: PChar; var PluginHandle: THandle): Integer;
var
  h: THandle;
  PluginVersion: array[0..2047] of Char;
  GetPluginInfo: TGetPluginInfo;
  IsSupported: TIsSupported;
begin
  PluginHandle := 0;
  Result := GF_UNKNOWN;
  h := LoadLibrary(PluginFileName);
  if h = 0 then
    Exit;
  try
    GetPluginInfo := GetProcAddress(h, 'GetPluginInfo');
    if not Assigned(GetPluginInfo) then
      Exit;
    GetPluginInfo(0, @PluginVersion, SizeOf(PluginVersion));
    IsSupported := GetProcAddress(h, 'IsSupported');
    if not Assigned(IsSupported) then
      Exit;
    if IsSupported(PChar(FileName), DWORD(pBuf)) <> 0 then
    begin
      PluginHandle := h;
      if PluginVersion = '00IN' then
        Result := GF_IMAGE
      else if PluginVersion = '00AM' then
        Result := GF_ARCHIVE
      else
        PluginHandle := 0;
      Exit;
    end;
  finally
    if PluginHandle = 0 then FreeLibrary(h);
  end;
end;

end.
